from AutoRPE.UtilsRPE.Error import LineNotFound, ProcedureNotFound
from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.TypeDealer import in_which_type_definition
from AutoRPE.UtilsRPE.Functions import remove_comments
import re
# This file includes tools to modify the code and make the implementation easier


# Get all files contained in a source path (recursive)
def expand_sources(source_path):
    from os import walk
    from os.path import join
    import re
    pattern = ".*\.[fFh90]"
    files = [join(dp, f) for dp, dn, fn in walk(source_path) for f in fn]
    files = [f for f in files if re.search(pattern, f)]
    # Sorting the list to solve the random-like behaviour of walk
    files = list(sorted(files))
    return files


# Gets the filepath of the corresponding module
def match_module_and_filename(module, filenames):
    import re
    pattern = ".*/%s\.[fFh90]" % module
    matches = [fn for fn in filenames if re.match(pattern, fn)]
    return matches


# Finds line at which the variable is declared
# WARNING: if the line of the declaration is split, it merges the line in the file
#          and therefore this variable can modify the file
def find_declaration_line(variable, filepath, vault, additional_files=None):
    import re
    from AutoRPE.UtilsRPE.Functions import remove_comments
    # Declaration patterns

    # Standard pattern
    pattern = r"::.*\b%s\b" % variable.name

    # Exception pattern (line is split)
    exception_pattern = r"&.*\b%s\b" % variable.name
    # Common real variable declaration pattern
    real_pattern = r"real.*::"

    def load_lines(filepath):
        with open(filepath, "r") as f:
            # Sometimes people decide that is ok to copy and paste non UTF-8 character in comments. It isn't.
            try:
                lines = [l for l in f]
            except UnicodeDecodeError:
                raise UnicodeDecodeError("The file %s contains non UTF-8 character.\n "
                                         "Please clean, and retry.", filepath)
        return lines

    # Check if the file contains exceptions and fix them
    check_for_exceptions = True
    while check_for_exceptions:
        # Setting this to False, will only activate this again in case a case is found
        check_for_exceptions = False
        lines = load_lines(filepath=filepath)
        current_block = CurrentBlock(vault[variable.routine.module.name].main())
        for index, line in enumerate(lines):
            line = remove_comments(line).strip()
            if current_block.routine == variable.routine:
                m = re.search(exception_pattern, line)
                if m:
                    declaration_displacement = 1
                    if remove_comments(lines[index - 1]).strip()[0] == "&":
                        while remove_comments(lines[index - declaration_displacement]).strip()[0] == "&":
                            declaration_displacement = declaration_displacement + 1
                    if re.search(real_pattern, lines[index - declaration_displacement], re.I):
                        merge_lines(filepath, index - declaration_displacement)
                        check_for_exceptions = True
                        break
            current_block = current_block.in_which_block(line, vault=vault)

    # Lines should be already updated since in case being modified it checks for exceptions again.

    # Search for declarations
    declarations = []
    current_block = CurrentBlock(vault[variable.routine.module.name].main())

    type_def = None
    for index, line in enumerate(lines):
        line = remove_comments(line).strip()
        type_def = in_which_type_definition(line, type_def)
        if re.search(pattern, line):
            if current_block.routine == variable.routine:
                if variable.is_member_of is None:
                    declarations.append(index)
                elif variable.is_member_of == type_def:
                    declarations.append(index)
                else:
                    pass
        else:
            current_block = current_block.in_which_block(line, vault=vault)
    if declarations:
        return declarations

    include_declaration = [line for line in lines if line.count("#include")]
    # If no line has been found in this file searches for possible additional files
    if additional_files is not None:
        for declaration in include_declaration:
            h90_file = declaration.split("#include")[1]
            if h90_file.lower().count("h90"):
                h90_file = h90_file.strip().replace("\"", "")
                additional_files.append(h90_file)
    raise LineNotFound


def merge_lines(file, index, fix_in_file=True):
    with open(file, "r") as f:
        lines = [l for l in f]
        line = remove_comments(lines[index]).strip()
        n_merged_lines = 1
        if line.count("&"):
            while line and line.strip()[-1] == "&":
                line1 = line
                while remove_comments(lines[index + n_merged_lines]).strip() == "":
                    n_merged_lines = n_merged_lines + 1
                line2 = remove_comments(lines[index + n_merged_lines])
                line = line1.replace("&", "").rstrip() + line2
                n_merged_lines = n_merged_lines + 1
            line = line.replace("&", "")
            line = re.sub(", +", ", ", line)
            line = re.sub(":  +", ":  ", line)
            line = line + "\n"
            if not fix_in_file:
                return line
        lines[index] = line
        for j in range(1, n_merged_lines):
            lines[index + j] = ""
    with open(file, "w") as f:
        text = "".join(lines)
        f.write(text)


# This function will find the file where a variable is declared and will replace the declaration
# with a new one using the new precision spec.
def replace_declaration(variable, file_paths, vault, target_precision="dp"):
    files = file_paths
    module = variable.routine.module.name
    module_file_paths = match_module_and_filename(module, files)
    replacement_done = False
    for filepath in module_file_paths:
        try:
            additional_files = []
            declaration_line_indexes = find_declaration_line(variable, filepath, vault, additional_files)
        except LineNotFound:
            if len(additional_files):
                for h90_file in additional_files:
                    h90_file_path = [file for file in file_paths if file.count(h90_file)]
                    module_file_paths.extend(h90_file_path)
            else:
                pass
            continue
        replacement_done = True
        with open(filepath, "r") as f:
            try:
                data = [l for l in f]
            except UnicodeDecodeError:
                raise UnicodeDecodeError("The file %s contains non UTF-8 character.\n "
                                         "Please clean, and retry.", filepath)
            for declaration_line_index in declaration_line_indexes:
                declaration_line = data[declaration_line_index]
                new_line = replace_real_declaration(variable, declaration_line, target_precision=target_precision,
                                                    filename=filepath)
                data[declaration_line_index] = new_line
            updated_text = "".join(data)
        with open(filepath, "w") as output:
            output.write(updated_text)
    if not replacement_done:
        print("Declaration of variable %s with id %i not found in module %s" % (
            variable.name, variable.id, variable.routine.module.name))

# This function takes a declaration line (string) and updates it with the new type
# If the declaration contains more variables it keeps the other variables with the original attributes.
def replace_real_declaration(variable, declaration_line, target_precision="dp", filename=None):
    from AutoRPE.UtilsRPE.Functions import remove_comments, split_declaration_arguments  # , split_declaration
    import re
    import warnings
    declaration_line_without_comments = remove_comments(declaration_line).strip()

    attributes, arguments = declaration_line_without_comments.split("::")
    # attributes = split_attributes(attributes)
    arguments = split_declaration_arguments(arguments)

    # Real statement pattern
    pattern = "real *\(([^\)]*)\)"

    if len(arguments) == 1:
        if declaration_line_without_comments[-1] == "&":
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
            print("WARNING!! the declaration of %s in %s hasn't been properly fixed!" % (variable.name, filename))
        else:
            m = re.search(pattern, declaration_line_without_comments, re.I)
            if m:
                real_statement = m.group(0)
                precision_specification = m.group(1)
                new_real_statement = real_statement.replace(precision_specification, target_precision)
                new_declaration_line_without_comments = \
                    declaration_line_without_comments.replace(real_statement, new_real_statement)
                new_declaration_line = \
                    declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
            else:
                warnings.warn("Problems finding real statement:\n%s" % declaration_line)
                new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                       declaration_line
        return new_declaration_line
    else:
        if variable.name not in arguments:
            raise AssertionError("Variable %s not found in list of arguments" % variable.name, attributes)
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
            leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line


def change_code(list_filename, source_path, vault_path, target_precision="dp"):
    # The purpose of this function is to modify the declarations of several variables in a source-code.
    # As inputs it takes the path to the file with the list of variables, the path to the sources that will be modified
    # and the path to the vault database containing information about the code.
    #
    # This function takes the following inputs:
    # - list_filename: Path to a file that contains a list of variables that will be modified in the code
    # - source_path: Path to a folder that contains the source files of the code corresponding to the vault.
    # - vault_path: Path to a vault database.
    from AutoRPE.UtilsRPE.Functions import load_vault, load_variable_list

    vault = load_vault(vault_path)
    variable_ids = load_variable_list(list_filename)
    file_paths = expand_sources(source_path=source_path)

    variables = [vault.get_variable_by_id(_id) for _id in variable_ids]
    for variable in variables:
        replace_declaration(variable, file_paths, vault, target_precision=target_precision)


def fix_casts(source_path, vault_path, list_filename=None):
    from AutoRPE.UtilsRPE.Functions import load_vault, calls_to_routines, find_call_arguments, \
        is_keyword_argument, close_brackets, \
        match_arguments_and_interface, load_variable_list
    from AutoRPE.UtilsRPE.BasicStructures import Interface
    from os.path import basename
    import re
    vault = load_vault(vault_path)
    pattern = r"call.*\b(real\(.*)"
    cast_pattern = r"\b(real\(.*)"
    pattern_precision = r"\b([wds]p|[8]) *\)"
    call_pattern = r"call *\b(\w*)\b"

    lookup_table = {52: "dp", 23: "sp"}

    if list_filename is not None:
        # Update vault precisions with variable list
        variables_to_update = load_variable_list(list_filename)
        for variable in vault.variables:
            if variable.id in variables_to_update:
                variable.precision = 52

    file_paths = expand_sources(source_path=source_path)
    for file_path in file_paths:
        if not (file_path.count(".F90") or file_path.count(".f90") or file_path.count(".h90")):
            continue
        file_modified = False
        module_name = basename(file_path).split(".")[0]
        with open(file_path, "r", encoding="latin-1") as f:
            lines = [l for l in f]
            for line_index, line in enumerate(lines):
                # Check if there's a routine call with a cast to real
                if re.search(pattern, line, re.I):
                    try:
                        calls = calls_to_routines(line)
                    except ValueError:
                        continue
                    for call in calls:
                        arguments = find_call_arguments(call)
                        routine_name = re.search(call_pattern, call, re.I).group(1)
                        try:
                            routine = vault.get_procedure_by_name(routine_name)
                        except ProcedureNotFound:
                            continue
                        # If it is an interface, find the specific routine being used
                        if isinstance(routine, Interface):
                            try:
                                routine = match_arguments_and_interface(arguments, interface=routine, vault=vault)
                            except AssertionError:
                                routine = None
                        for index, argument in enumerate(arguments):
                            if re.search(cast_pattern, argument, re.I):
                                if routine is not None:
                                    if not is_keyword_argument(argument):
                                        target_precision = routine.get_variable_by_position(index).precision
                                    else:
                                        key = argument.split("=")[0].strip()
                                        target_precision = routine.get_variable_by_name(key).precision
                                else:
                                    # For now we will assume that in this case it is not necessary to change the cast
                                    continue
                                    # Problematic to assume that the precision of an external routine is either double or single
                                    # target_precision = 52
                                    target_precision = 23
                                try:
                                    target_precision = lookup_table[target_precision]
                                except KeyError:
                                    raise KeyError("target_precision %s not found in lookup_table", target_precision)
                                cast = close_brackets(re.search(cast_pattern, argument, re.I).group(1))
                                precision = re.search(pattern_precision, cast, re.I).group(1)
                                if precision.strip() == target_precision:
                                    continue
                                else:
                                    cast_replacement = re.sub(r"(.*\b)[wds]p( *\).*)", r"\1%s\2" % target_precision,
                                                              cast, re.I)
                                    line = line.replace(cast, cast_replacement)
                                    lines[line_index] = line
                                    file_modified = True
        if file_modified:
            text = "".join(lines).strip()
            with open(file_path, "w") as f:
                f.write(text)

# Small function created to easily find the id of variables that appear in the code
def get_variable_id(vault, variable_name, routine_name=None, module_name=None, ):
    vs = [v for v in vault.variables if v.name == variable_name]

    if module_name is not None:
        vs = [v for v in vs if v.routine.module.name == module_name]
    if routine_name is not None:
        vs = [v for v in vs if v.routine.name == routine_name]

    if len(vs) == 1:
        print("Variable found! ID: %i" % vs[0].id)
        return vs[0].id
    elif len(vs) > 1:
        print("More than one variable!")
        return None
    else:
        print("Not a single variable found!")
        return None


# The purpose of this function is to assign ids to rpe_variables that didn't have any assigned.
def expand_vault(vault):
    rpe_vars = [v for v in vault.variables if v.type == "rpe_var"]
    for var in rpe_vars:
        var.mutable = True
    rpe_vars_with_id = [v for v in rpe_vars if v.id is not None]
    rpe_vars_without_id = [v for v in rpe_vars if v.id is None]

    last_id = max([v.id for v in rpe_vars_with_id])

    counter = last_id + 1
    for v in rpe_vars_without_id:
        v.id = counter
        v.mutable = False
        counter += 1


# Few test cases
def case_update_vault_with_precision_from_source():
    source_path = "/home/Earth/otintopr/Workspace/test_codechange/src/OCE"
    vault_path = "/esarchive/scratch/otinto/ExperimentAutoRPE_45days/ConfigurationFiles/vault-NEMO4.0.1.pkl"

    lookup_table = {'WP': 23, 'default': 23, '8': 52, 'dp': 52, 'sp': 23, 'fbsp': 23, 'wp': 23}
    vault = update_vault_with_precision_from_source(source_path=source_path, vault_path=vault_path,
                                                    lookup_table=lookup_table)

    not_found_vars = [v for v in vault.variables if v.id is not None and v.precision is None]
    print(len(not_found_vars))


def case_expand_vault():
    from AutoRPE.UtilsRPE.Functions import load_vault
    vault_path = "/esarchive/scratch/otinto/ExperimentAutoRPE_45days/ConfigurationFiles/vault-NEMO4.0.1.pkl"

    vault = load_vault(vault_path)

    expand_vault(vault)

    get_variable_id(vault, "pts", module_name="eosbn2", routine_name="eos_insitu")


def case_expand_plus_update_vault_with_precision():
    from AutoRPE.UtilsRPE.Functions import save_variable_list
    source_path = "/home/Earth/otintopr/Workspace/test_codechange/src/OCE"
    vault_path = "/esarchive/scratch/otinto/ExperimentAutoRPE_45days/ConfigurationFiles/vault-NEMO4.0.1.pkl"

    # Define which number of significant bits correspond to each type specification found in the code
    lookup_table = {'WP': 23, 'default': 23, '8': 52, 'dp': 52, 'sp': 23, 'fbsp': 23, 'wp': 23}

    # Read the precision status from source
    vault = update_vault_with_precision_from_source(source_path=source_path, vault_path=vault_path,
                                                    lookup_table=lookup_table)

    # Assign ids to variables that doesn't have one
    expand_vault(vault)

    # Look for the variables that are in double precision in the source and find its ids
    double_precision_variables = [v for v in vault.variables if v.precision == 52]
    variable_ids = [v.id for v in double_precision_variables]

    # Save a list with these variables
    save_variable_list(list_of_variables=variable_ids, filename="/home/Earth/otintopr/test_dp_vars.txt", vault=vault)

    vault.generate_namelist_precisions(filename="/home/Earth/otintopr/test_dp_vars_namelist.txt",
                                       exclude_double_precision=False)


if __name__ == "__main__":
    case_expand_plus_update_vault_with_precision()
